/***************************************************************************
 *   $Id: csvData.c,v 1.5 2007/02/20 10:11:40 kenjen Exp $
 *                                                                         *
 *   Created: Wed Jun 5 2002
 *                                                                         *
 *   Copyright (C) 2002-2005 by Kenneth Jennings                           *
 *   ken_jennings@bellsouth.net                                            *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <csvData.h>

static char rcsid[] = "$Id: csvData.c,v 1.5 2007/02/20 10:11:40 kenjen Exp $";

char * getRCSIDcsvData( void ) { return rcsid; }

typedef struct _csvtotext_
{
	unsigned long bit;
	char *        text;
} CsvToText;


static CsvToText csvToText[] =
{
	{ CSV_NO_VALIDATION, "CSV_NO_VALIDATION" }, /* Note this "bit" is ZERO! */
	{ CSV_NULL,          "CSV_NULL"          },
	{ CSV_OVERRUN,       "CSV_OVERRUN"       },
	{ CSV_MAXSIZE,       "CSV_MAXSIZE"       },
	{ CSV_MINSIZE,       "CSV_MINSIZE"       },
	{ CSV_ALPHAUPPER,    "CSV_ALPHAUPPER"    },
	{ CSV_ALPHALOWER,    "CSV_ALPHALOWER"    },
	{ CSV_DIGITS,        "CSV_DIGITS"        },
	{ CSV_SPACES,        "CSV_SPACES"        },
	{ CSV_MINUS,         "CSV_MINUS"         },
	{ CSV_PLUS,          "CSV_PLUS"          },
	{ CSV_PARENS,        "CSV_PARENS"        },
	{ CSV_BRACKETS,      "CSV_BRACKETS"      },
	{ CSV_AMPERSAND,     "CSV_AMPERSAND"     },
	{ CSV_APOSTROPHE,    "CSV_APOSTROPHE"    },
	{ CSV_PERIOD,        "CSV_PERIOD"        },
	{ CSV_SLASH,         "CSV_SLASH"         },
	{ CSV_COMMA,         "CSV_COMMA"         },
	{ CSV_DOLLAR,        "CSV_DOLLAR"        },
	{ CSV_OTHER_SYMBOL,  "CSV_OTHER_SYMBOL"  },
	{ 0L,                "UNKNOWN"           }
};


/*
 **********************************************************
 *
 * P R I V A T E   F U N C T I O N   P R O T O T Y P E S
 *
 **********************************************************
 */

static char * csvReportRuleAsText(unsigned long bits);

static long csvSanityCheck(CsvData * field);

static long csvAllocFieldData(CsvData * field);

static long csvCleanField(CsvData * field);

static long csvCheckEndOfField(CsvData * field);

static long csvAddChar(char      thisChar,
					   CsvData * field);

static int  csvCheckValidChar(int           inProcess,
							  char          min,
							  char          max,
							  unsigned long test,
							  char          data,
							  CsvData *     field);

static int  csvCheckValid(char      data,
						  CsvData * field);


/*
 **********************************************************
 *
 * P R I V A T E   F U N C T I O N S
 *
 **********************************************************
 */

/* Report the text name for a given CSV_BITMAP attribute.
 * Used to report what fields have an error, though it
 * could be used for any diagnostics.
 * Note that only the first bit matched is returned.
 */

static char * csvReportRuleAsText(unsigned long bits)
{
	int index = 1; /* Note 1, not 0.  Bit value in index 0 is 0 for NO_VALIDATION */

	if (bits == 0)
		return csvToText[0].text;

	while (csvToText[index].bit)
	{
		if (csvToText[index].bit & bits)
			return csvToText[index].text;

		index++;
	}

	return csvToText[index].text; /* Last one in list is UNKNOWN */
}

/*
 * Validate that the input values in the struct are coherent.
 */

static long csvSanityCheck(CsvData * field)
{
	unsigned long retVal = CSV_OK;


	if (!field->buf &&
		 field->aloc) /* no pointer, but allocation flag set */
		return CSV_DATA_STRUCT;

	if ((field->minLen > field->maxLen) ||
		(field->minLen < 0))
		return CSV_DATA_STRUCT;

	if (!field->buf)
		retVal = csvAllocFieldData(field); /* make space if necessary for parsed data */

	return retVal;
}



/*
 * Allocate a buffer for this field if no allocation
 * has been done.
 */

static long csvAllocFieldData(CsvData * field)
{
	if (field->buf  ||        /* there already is a buf */
		field->aloc ||        /* already flagged as allocated */
		(field->maxLen < 1) ) /* allocation length is bogus */
		return CSV_DATA_STRUCT;

	field->buf = (char *)malloc(field->maxLen + 1);

	if (!field->buf)
		return CSV_MEMORY;

	field->buf[0] = '\0';
	field->aloc   = true;

	return CSV_OK;
}

int freeCsvData(CsvData * colSpec)
{
	if (!colSpec)
		return -1;

	if (colSpec->aloc &&
		colSpec->buf)
		free(colSpec->buf );

	free(colSpec);

	return 0;
}


CsvData * createCsvData(void)
{
	CsvData * newData;

	newData = (CsvData *)malloc(sizeof(CsvData));

	if (!newData)
		return NULL;

	newData->buf    = NULL;
	newData->aloc   = false;
	newData->len    = 0;
	newData->fail   = 0L;
	newData->rule   = CSV_EVERYTHING | CSV_NULL | CSV_OVERRUN;
	newData->minLen = 0;
	newData->maxLen = 256;

	return newData;
}


/*
 * Set all field values to begin processing.
 */

static long csvCleanField(CsvData * field)
{
	field->len  = 0;
	field->fail = 0L;

	if (!field->buf)
		return CSV_DATA_STRUCT;

	field->buf[0] = '\0';

	if (field->maxLen < 1)
		return CSV_DATA_STRUCT;

	field->buf[field->maxLen] = '\0';

	return CSV_OK;
}



/*
 * This is the end of the field. Check that the lengths
 * meet validation criteria.
 */

static long csvCheckEndOfField(CsvData * field)
{
	if (field->rule == CSV_NO_VALIDATION)
		return CSV_OK;

	if (field->len == 0)                /* An empty string */
	{
		if (!(field->rule & CSV_NULL) ) /* And NULL is not allowed? */
		{
			field->fail |= CSV_NULL;

			return CSV_VALIDATION;
		}
	}

	if (field->len < field->minLen)     /* Too short for MIN len */
	{
		if (field->rule & CSV_MINSIZE ) /* And MINSIZE is required? */
		{
			field->fail |= CSV_MINSIZE;

			return CSV_VALIDATION;
		}
	}

	if (field->len < field->maxLen)     /* Too short for MAX len */
	{
		if (field->rule & CSV_MAXSIZE ) /* And MAXSIZE is required? */
		{
			field->fail |= CSV_MAXSIZE;

			return CSV_VALIDATION;
		}
	}

	return CSV_OK;
}



/*
 * Add a character to this field.
 */

static long csvAddChar(char thisChar,
					   CsvData * field)
{
	if (field->len == field->maxLen) /* Already reached max?, so this is overrun. */
	{
		if (!( (field->rule == CSV_NO_VALIDATION) || /* But Overrun is not allowed? */
			   (field->rule &  CSV_OVERRUN) ) )
		{
			field->fail |= CSV_OVERRUN;

			return CSV_VALIDATION;
		}
	}

	if (field->len < field->maxLen)
	{
		field->buf[field->len]     = thisChar;
		field->buf[field->len + 1] = '\0';
	}

	field->len++; /* count every character, even after overrun */

	return CSV_OK;
}



/*
 * Validate if this char is allowed in this field
 */

static int csvCheckValidChar(int           inProcess,
							 char          min,
							 char          max,
							 unsigned long test,
							 char          data,
							 CsvData *     field)
{
	if ((inProcess == CSV_NOT_DECIDED) &&
		(data >= min) &&
		(data <= max))
	{
		if (field->rule & test)
			inProcess = CSV_APPROVED;
		else
		{
			inProcess = CSV_DISAPPROVED;
			field->fail |= test;
		}
	}

	return inProcess;
}



/*
 * Validate all the allowable char possibilites for the given
 * data and the field.
 */

static int csvCheckValid(char      data,
						 CsvData * field)
{
	int inProcess = CSV_NOT_DECIDED;
	int loop      = 0;

	struct _cvc_
	{
		char          min;
		char          max;
		unsigned long test;
	} ckvc[] =
	{
		{ 'A',  'Z',  CSV_ALPHAUPPER },
		{ 'a',  'z',  CSV_ALPHALOWER },
		{ '0',  '9',  CSV_DIGITS     },
		{ ' ',  ' ',  CSV_SPACES     },
		{ '-',  '-',  CSV_MINUS      },
		{ '+',  '+',  CSV_PLUS       },
		{ '(',  ')',  CSV_PARENS     },
		{ '[',  '[',  CSV_BRACKETS   },
		{ ']',  ']',  CSV_BRACKETS   },
		{ '&',  '&',  CSV_AMPERSAND  },
		{ '\'', '\'', CSV_APOSTROPHE },
		{ '.',  '.',  CSV_PERIOD     },
		{ '/',  '/',  CSV_SLASH      },
		{ ',',  ',',  CSV_COMMA      },
		{ '$',  '$',  CSV_DOLLAR     },
		{ '\0', '\0', 0L             }
	};


	if (field->rule == CSV_NO_VALIDATION) /* Allow everything */
		return CSV_APPROVED;
	
	while (ckvc[loop].test)
	{
		inProcess = csvCheckValidChar(inProcess,       /* inProcess */
									  ckvc[loop].min,  /* from      */
									  ckvc[loop].max,  /* to        */
									  ckvc[loop].test, /* bit map   */
									  data,            /* data      */
									  field);          /* field     */

		loop++;
	}

	if (inProcess == CSV_NOT_DECIDED) /* for all other symbols */
	{
		if (field->rule & CSV_OTHER_SYMBOL)
			inProcess = CSV_APPROVED;
		else
		{
			inProcess = CSV_DISAPPROVED;
			field->fail |= CSV_OTHER_SYMBOL;
		}
	}

	return inProcess;
}



/*
 **********************************************************
 *
 * P U B L I C   F U N C T I O N S
 *
 **********************************************************
 */

long csvParse(char    *  data,   /* pointer to char data to parse */
			  CsvData ** fields, /* NULL terminated list of pointers to field structures  */
			  char       sv)     /* separator character between fields */
{
	CsvData * thisField = NULL;
	char      thisChar;
	long      retVal    = CSV_OK;
	int       inProcess;
	bool      quotedFlag;
	bool      firstChar;

	/* Eliminate the possibilities of null pointers, null characters */

	if (!data   || !*data   || /* NULL data pointer, or blank string */
		!fields || !*fields || /* NULL fields list, or first list item NULL */
		!sv)                   /* '\0' separator character. */
		return CSV_PARAMETER;


	while (*fields && (retVal == CSV_OK)) /* loop as long as there are fields and everything's OK */
	{
		quotedFlag = false;

		thisField = *fields;

		retVal = csvSanityCheck(thisField); /* Make sure structure is coherent */

		if (retVal != CSV_OK)
			continue;

		retVal = csvCleanField(thisField); /* Init 0, null, blank values */

		if (retVal != CSV_OK)
			continue;

		thisChar = *data;

		firstChar = true;

		while (thisChar && (retVal == CSV_OK)) /* loop while there is data and everything's OK */
		{
			if ((firstChar) && (thisChar == '\"'))
			{
				quotedFlag = true;
				firstChar = false;
				data++;
				thisChar = *data;
				continue;
			}

			if ((quotedFlag) && (thisChar == '\"'))
			{
				quotedFlag = false;
				data++;
				thisChar = *data;
				continue;
			}

			if (((thisChar == sv) && (!quotedFlag)) ||
				(thisChar == CSV_LF) || /* End of line chars are same */
				(thisChar == CSV_CR))   /* as end of string to parse  */
			{
				thisChar = '\0';

				continue;
			}

			inProcess = csvCheckValid(thisChar,
									  thisField);

			/* If it failed validation then there's nothing more to try */

			if (inProcess == CSV_DISAPPROVED)
				retVal = CSV_VALIDATION; /* fail bits set in csvCheckValidChar() */

			if (inProcess == CSV_APPROVED)
			{
				retVal = csvAddChar(thisChar,
									thisField); /* Add with overrun detection */
			}

			data++;
			thisChar = *data;
			firstChar = false;
		}

		if (!thisChar && (retVal == CSV_OK)) /* Between fields Check lengths */
			retVal = csvCheckEndOfField(thisField);

		if ((*data == sv) && (retVal == CSV_OK)) /* expect another field, so */
			data++;                              /* need to skip over separator */

		if (retVal == CSV_OK)
			fields++;
	}

	if ((retVal == CSV_OK)        && /* The parsing was successful */
		(*fields                  || /* There is still a field OR */
		 (*data                   &&  /* There is still data AND */
		  !((*data == CSV_LF)     ||  /* remaining data is not either */
		    (*data == CSV_CR) ) ) ) ) /* end of line character AND */
	{
		if (*fields) /* more field structs than data to parse */
			retVal = CSV_EXCESS_STRUCTS;
		else         /* more data to parse than structs to fill */
			retVal = CSV_EXCESS_DATA;
	}

	return retVal;
}



/* Convert csvParse return value to text */

char * csvRetValToText(long retVal)
{
	struct _csvretval_
	{
		 long  retVal;
		char * text;
	} csvRetVal[] =
	{
		{ CSV_OK,             "CSV_OK"             }, /* Note this "bit" is ZERO! */
		{ CSV_PARAMETER,      "CSV_PARAMETER"      },
		{ CSV_EXCESS_STRUCTS, "CSV_EXCESS_STRUCTS" },
		{ CSV_EXCESS_DATA,    "CSV_EXCESS_DATA"    },
		{ CSV_DATA_STRUCT,    "CSV_DATA_STRUCT"    },
		{ CSV_MEMORY,         "CSV_MEMORY"         },
		{ CSV_VALIDATION,     "CSV_VALIDATION"     },
		{ 0L,                 "UNKNOWN"            },
		{ 0L,                 NULL                 }
	};

	int index = 0;

	while (csvRetVal[index].text)
	{
		if (csvRetVal[index].retVal == retVal)
			return csvRetVal[index].text;

		index++;
	}

	index--; /* We reached the NULL, back up one entry */
	return csvRetVal[index].text; /* Last one in list is UNKNOWN */
}



/* Convert a text string to the appropriate rule(s) bit(s).
 * Note that since strstr() is used to search, the string
 * supplied to the function can contain multiple CSV_* value
 * names provided they are delimited somehow (commas, slashes,
 * vertical bars, or spaces.)
 * CSV_NO_VALIDATION is checked for first.  If it is found
 * it overrides all other bits.
 */

unsigned long csvTextToRule(char * text)
{
	int           index = 1;  /* Note 1, not 0.  Bit value in index 0 is 0 for NO_VALIDATION */
	unsigned long bits  = 0L;


	if (!text || !*text)
		return CSV_NO_VALIDATION;

	if (strstr(text, csvToText[0].text)) /* Check for CSV_NO_VALIDATION first */
		return CSV_NO_VALIDATION;

	while (csvToText[index].bit)
	{
		if (strstr(text, csvToText[0].text))
			bits |= csvToText[0].bit;

		index++;
	}

	return bits;
}



/* Report first known error as a text message for visible diagnostics */

char * csvDiag(CsvData ** fields)
{
	static char message[200];
	CsvData *   thisField;
	int         index;

	message[0] = '\0';
	index      = 0;

	if (!fields)
		return NULL;

	while (*fields)
	{
		thisField = *fields;

		if (thisField->fail)
		{
			sprintf(message,
					"CsvData field %d failed to extract data.  Rule violated: %s",
					index + 1,
					csvReportRuleAsText(thisField->fail));
			break;
		}

		index++;
		fields++;
	}

	return message;
}



long csvCleanUp(CsvData ** fields) /* NULL terminated list of pointers to field structures  */
{
	CsvData * thisField;

	/* I opted to excluded the check for !*fields which would have
	 * caused the CSV_PARAMETER return value if the first item on the
	 * list is NULL.  The rest of the function's logic deals with that
	 * NULL appropriately, so I let it pass here.
	 */

	if (!fields)
		return CSV_PARAMETER;

	while (*fields)
	{
		thisField = *fields;

		if (thisField->aloc &&
			thisField->buf)
		{
			free(thisField->buf);

			thisField->buf  = NULL;
			thisField->aloc = false;
		}

		fields++;
	}

	return CSV_OK;
}


/* re-sprintf and realloc buffer */
/* Pay attention - the internally declared size of
 * tStr may possibly cause a crash if the caller
 * uses an unreasonable format for the string. */

int reformatCsvDataLong(CsvData * field,
						char    * fmt)
{
	char   tStr[30];
	long   tLong;
	int    tLen;
	char * tCpy;

	if (!field || !field->buf ||
		!fmt   || !*fmt)
		return -1;

	tLong = atol(field->buf);

	tStr[0] = '\0';

	if (!sprintf(tStr, fmt, tLong) )
		return -1;

	tLen = strlen(tStr);

	if (!tLen || (tLen > field->maxLen) )
		return -1;

	tCpy = (char *)malloc(tLen + 1);

	if (!tCpy)
		return -1;

	strcpy(tCpy, tStr);

	free(field->buf);
	field->buf = tCpy;

	return 0;
}


