/***************************************************************************
 *   $Id: csvData.h,v 1.4 2005/02/27 04:26:53 kenjen Exp $ *
 *                                                                         *
 *   Created: Wed Jun 5 2002                                               *
 *                                                                         *
 *   Copyright (C) 2002-2005 by Kenneth Jennings                           *
 *   ken_jennings@bellsouth.net                                            *
 *                                                                         *
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

 /*
 * Include file for generic handling of CSV (comma separated
 * value) data.  This is typically found in databases and
 * spreadsheets that have been saved/dumped in an ASCII
 * format.
 *
 * You supply the data structures representing each field
 * and a pointer to the buffer read from the file and
 * then this code will fill in the structures until it
 * runs out of buffer data or runs out of data structures.
 *
 * A line of data must have the same number of fields as the
 * number of data structures to be filled in.  More data to parse
 * than structures or more structures than data to parse will
 * cause the parsing to stop and an error be returned.
 *
 * While this is "Comma" Separated Values, the Comma isn't a
 * required separator.  The call to parse now includes the
 * character used to separate data.  Use vertical bars '|',
 * or slashes '/' or whatever makes you happy.
 *
 ********************************************************************
 *
 * This is rather quick-n-dirty, so not everything documented here
 * may be working (or bullet proof).
 *
 * Quoting is now being handled.  If the first character of a field
 * is a double quote " then all characters are assumed to be part
 * of the field until the closing double quote.  Note that the quotes
 * are stripped automatically and are not considered part of the
 * field data.
 *
 * Leading and trailing spaces are considered data.  Be careful of
 * extra blank spaces at the end of lines.  If you don't want
 * trailing spaces, then deblank the string to be parsed prior to
 * calling csvParse();
 *
 ********************************************************************
 *
 * Not entirely obvious:
 *
 * If you specify CSV_NO_VALIDATION in a field any number of
 * characters may be in the file data for the field, including
 * no characters.  Thus CSV_NO_VALIDATION is the same thing
 * as ALL other options turned on.  This value (0) is checked
 * for before ALL other processing options.  Ultimately, this means
 * that a field list containing one field will correctly parse
 * a blank line. If this is not what you want then use validation.
 *
 * CSV_NO_VALIDATION does not superceed the check for the
 * correct number of fields in the data.  The correct number
 * of separator characters must be in the data.
 *
 * Without CSV_NO_VALIDATION data overrun is allowed only if
 * CSV_OVERRUN is specified.  Likewise, NULL values are only
 * allowed if CSV_NULL is specified.  By default, some kind
 * of data is expected in a field if ANY validation bit is
 * set.  Remember to use CSV_NULL if empty fields are possible.
 *
 * The fail value in the structure returns the validation
 * approval setting that was violated.  Thus, if you do not
 * specify a given validation bit approving a character, that
 * bit will be a reason for failure.
 *
 * If You do not specify CSV_NULL, then NULL fields are not
 * allowed, so CSV_NULL may be a failure bit set. etc.
 * CSV_OTHER_SYMBOL allows any character OTHER than the ones not
 * specifically listed in the validation.  Therefore, if
 * CSV_OTHER_SYMBOL is not set, then this will be the fail bit
 * set IF a previous validation check did not approve the
 * cadidate character first.  Note that CSV_OTHER_SYMBOL does
 * not imply all possible characters, just the ones that are
 * not previously on the approval/disapproval list.
 *
 * The validation for CSV_NULL comes before the validation for
 * CSV_MINSIZE.  This means that setting a minLen value of 0 is not
 * the same as allowing a NULL value.  If you want to allow NULL
 * values you should not set the CSV_MINSIZE bit at all.
 *
 * Sanity checks are done on the function parameters and the
 * field before processing the csv data begins.  The list of
 * return values are all greater than CSV_OK (0) and arrange all
 * the sanity check values before the value for validation.  So,
 * CSV_OK is ALWAYS the smallest return value, and CSV_VALIDATION
 * is ALWAYS the largest.
 *
 * So, what this means:
 *
 * retVal = csvParse(thisthat, andtheother, ',');
 *
 * if ((retVal >  CSV_OK) &&
 *     (retVal <  CSV_VALIDATION))
 *     printf("Structure/Memory/Functional problem");
 *
 * if (retVal == CSV_VALIDATION) // always the last code
 * {
 *     blah blah,
 *     figure out which struct has fail bits
 * }
 *
 ********************************************************************
 *
 * NOTE that if any return value other than CSV_OK is returned it
 * means that processing stopped in a field.  Therefore no data
 * in any field structure should be trusted.  If CSV_VALIDATION
 * is returned than the fail item of the structs can be examined
 * to determine where processing stopped.  Additionally, none
 * of the data in structures on the field list following the first
 * one with fail bits set is to be trusted, since processing stops
 * at the item flagged by fail bits.
 *
 ********************************************************************
 *
 * There are two ways of specifying data buffers for fields:
 * 1) declare the buffer yourself and supply the pointer as buf.
 * 2) leave the buf item NULL and the code will automagically
 *    allocate space to (maxLen + 1).  In this case, when you're
 *    done with the structures you must call csvCleanUp() to
 *    free that allocated space.
 * MAKE SURE THAT *ALL* FIELD STRUCTURES ARE DECLARED WITH THE
 * aloc == false!!!!  This value decides whether or not the buffers
 * were dynamically allocated which determines if the code attempts
 * to free(buf).  Baaaad things happen if you try to free a declared
 * array as if it were malloc()'d memory.
 *
 ********************************************************************
 *
 * TO DO:
 *
 ********************************************************************
 *
 * First modification: The separator need not be a comma.  It
 * was a simple mod to open the parser up to key on any given
 * separator character, so this is now a parameter to the
 * parsing function.
 *
 ********************************************************************
 *
 * Example of parsing two fields:
 *
 * char field1buf[41];
 *
 * CsvData field1 =
 * {
 *    field1buf, false, 0, 0L, // Output data
 *    CSV_NO_VALIDATION,
 *    0, 40
 * };
 *
 * CsvData field2 =
 * {
 *    NULL, false, 0, 0L, // Output data
 *    CSV_OVERRUN | CSV_PRESENT | CSV_ALL_CHARS,
 *    10, 15 };
 *
 * CsvData * fields[] =
 * {
 *    &field1,
 *    &field2,
 *    NULL
 * };
 *
 * //blah blah, open a file, read data into  char lineData[];
 *
 * long retVal;
 *
 * retVal = csvParse(lineData,
 *                   fields,
 *                   ',');
 *
 * if (retVal == CSV_OK)
 * {
 *    // do whatever with field1.buf (field1buf[]) and
 *    // field2.buf (alloc'd memory)
 *    // etc.
 * }
 *
 * csvCleanUp(fields); // NECESSARY, because csvParse() allocated memory!!!
 */


#ifndef CSV_DATA_STRUCT_INCL
#define CSV_DATA_STRUCT_INCL

#include <csvType.h>


/* Bit-maps giving the validation flags */

#define CSV_NO_VALIDATION 0x00000000 /* EXCLUSIVE of the values below! */

#define CSV_NULL          0x00000001 /* Allows no data at all */
#define CSV_OVERRUN       0x00000002 /* Allows more chars than fit in buf */

#define CSV_MAXSIZE       0x00000004 /* Must fill to maxLen */
#define CSV_MINSIZE       0x00000008 /* Must fill to at least minLen */

#define CSV_ALPHAUPPER    0x00000010 /* Allows 'A' through 'Z' */
#define CSV_ALPHALOWER    0x00000020 /* Allows 'a' through 'z' */
#define CSV_DIGITS        0x00000040 /* Allows '0' thorugh '9' */
#define CSV_SPACES        0x00000080 /* Allows blank spaces ' '   */
#define CSV_MINUS         0x00000100 /* Allows minus/hyphen '-'   */
#define CSV_PLUS          0x00000200 /* Allows plus         '+'   */
#define CSV_PARENS        0x00000400 /* Allows parenthesis  '( )' */
#define CSV_BRACKETS      0x00000800 /* Allows brackets     '[ ]' */
#define CSV_AMPERSAND     0x00001000 /* Allows ampersand    '&'   */
#define CSV_APOSTROPHE    0x00002000 /* Allows single quote '     */
#define CSV_PERIOD        0x00004000 /* Allows a period     '.'   */
#define CSV_SLASH         0x00008000 /* Allows a slash      '/'   */
#define CSV_COMMA         0x00010000 /* Allows a comma      ','   */
#define CSV_DOLLAR        0x00020000 /* Allows a dollar     '$'   */
#define CSV_OTHER_SYMBOL  0x10000000 /* Allows EVERY character */

#define CSV_PRESENT    CSV_MINSIZE  /* Used to be CSV_MAXSIZE | CSV_MINSIZE */

#define CSV_ALPHA      CSV_ALPHAUPPER | CSV_ALPHALOWER
#define CSV_SIGN       CSV_MINUS | CSV_PLUS
#define CSV_SYMBOLS    CSV_SIGN | CSV_PARENS | CSV_BRACKETS | CSV_AMPERSAND | CSV_APOSTROPHE | CSV_PERIOD | CSV_SLASH | CSV_COMMA | CSV_DOLLAR
#define CSV_ALL_CHARS  CSV_ALPHA | CSV_DIGITS | CSV_SYMBOLS
#define CSV_EVERYTHING CSV_ALL_CHARS | CSV_SPACES | CSV_OTHER_SYMBOL

#define CSV_AMOUNT     CSV_DIGITS | CSV_PERIOD | CSV_SIGN

#define CSV_DATE1      CSV_DIGITS | CSV_SLASH /* M/D = 1/1 or 12/31 */

#define CSV_MONEY1     CSV_DIGITS | CSV_PERIOD | CSV_COMMA | CSV_DOLLAR | CSV_PARENS
#define CSV_MONEY2     CSV_DIGITS | CSV_PERIOD | CSV_COMMA | CSV_DOLLAR | CSV_PARENS | CSV_SIGN

#define CSV_FP1        CSV_DIGITS | CSV_PERIOD
#define CSV_FP2        CSV_DIGITS | CSV_PERIOD | CSV_SIGN



/* A shortcut for initializing the first four
 * data items in the structures.
 */

#define CSV_INIT NULL, false, 0, 0L

/* Return values from the functions : */

#define CSV_OK             0 /* Does this really need explaination??? */

#define CSV_PARAMETER      1 /* A function parameter is NULL or nonsense */
#define CSV_EXCESS_STRUCTS 2 /* More structures than data */
#define CSV_EXCESS_DATA    3 /* More file data than structures */
#define CSV_DATA_STRUCT    4 /* dumb data in struct: 0 maxLen?, etc. */
#define CSV_MEMORY         5 /* Failed to allocate or free memory */

#define CSV_VALIDATION     6 /* Always last: A field failed validation.  see fail value */


/* Character processing during parsing*/

#define CSV_APPROVED       1 /* character approved during parsing */
#define CSV_DISAPPROVED    2 /* character disapproved during parsing */
#define CSV_NOT_DECIDED    3 /* we're not done testing this yet. */

/* The structure representing a field value parsed from the file */

typedef struct
{
	char *        buf;    /* Output: Where the data ends up */
	bool          aloc;   /* Output: flag if csvParse() allocated memory for buf */
	int           len;    /* Output: Current length of data */
	unsigned long fail;   /* Output: Validation rule that was broken */
	unsigned long rule;   /* Input:  The CSV_#? validation bits above */
	int           minLen; /* Input:  The minumum length of data */
	int           maxLen; /* Input:  The maximum length of data */
} CsvData;


/* low level code for csvFile() */

int freeCsvData(CsvData * colSpec); /* dump a dynamically allocated CsvData */

CsvData * createCsvData(void); /* Dynamically allocate a generic CsvData */

int reformatCsvDataLong(CsvData * field,
						char    * fmt); /* re-sprintf and realloc buffer */

/* user code */

long csvParse(char    *  data,   /* pointer to char data to parse */
			  CsvData ** fields, /* NULL terminated list of pointers to field structures  */
			  char       sv);    /* separator character between fields */

unsigned long csvTextToRule(char * text); /* Convert CSV_* text names to a CSV_* bit map */

char * csvRetValToText(long retVal); /* Convert csvParse return value to text */

char * csvDiag(CsvData ** fields); /* Report first known error as a text message for visible diagnostics */

long csvCleanUp(CsvData ** fields); /* NULL terminated list of pointers to field structures  */

#endif

